"""The stats sub-package contains statistical methods for Pastas.

.. currentmodule:: pastas.stats

.. autosummary::
    :toctree: ./generated
    :nosignatures:

    core
    metrics
    dutch
    sgi
    signatures
    tests

"""

# flake8: noqa

import pastas.stats.metrics as metrics
import pastas.stats.signatures as signatures

from .core import acf, ccf, mean, std, var
from .dutch import ghg, glg, gvg, q_ghg, q_glg, q_gvg
from .metrics import aic, bic, evp, kge, kge_2012, mae, nse, pearsonr, rmse, rsq, sse
from .sgi import sgi
from .tests import diagnostics, durbin_watson, ljung_box, runs_test, stoffer_toloi



"""This module contains utility functions for plotting."""

from typing import List, Union

import numpy as np
from pandas import Series

from pastas.typing import Axes


def _table_formatter_params(s: float, na_rep: str = "") -> str:
    """Internal method for formatting parameters in tables in Pastas plots.

    Parameters
    ----------
    s : float
        value to format.

    Returns
    -------
    str
        float formatted as str.
    """
    if np.isnan(s):
        return na_rep
    elif np.floor(np.log10(np.abs(s))) <= -2:
        return f"{s:.2e}"
    elif np.floor(np.log10(np.abs(s))) > 5:
        return f"{s:.2e}"
    else:
        return f"{s:.2f}"


def _table_formatter_stderr(s: float, na_rep: str = "") -> str:
    """Internal method for formatting stderrs in tables in Pastas plots.

    Parameters
    ----------
    s : float
        value to format.

    Returns
    -------
    str
        float formatted as str.
    """
    if np.isnan(s):
        return na_rep
    elif np.floor(np.log10(np.abs(s))) <= -4:
        return f"±{s * 100.:.2e}%"
    elif np.floor(np.log10(np.abs(s))) > 3:
        return f"±{s * 100.:.2e}%"
    else:
        return f"±{s:.2%}"


def _get_height_ratios(ylims: List[Union[list, tuple]]) -> List[float]:
    height_ratios = []
    for ylim in ylims:
        hr = ylim[1] - ylim[0]
        if np.isnan(hr):
            hr = 0.0
        height_ratios.append(hr)
    return height_ratios


def _get_stress_series(ml, split: bool = True) -> List[Series]:
    stresses = []
    for name in ml.stressmodels.keys():
        nstress = len(ml.stressmodels[name].stress)
        if split and nstress > 1:
            for istress in range(nstress):
                stress = ml.get_stress(name, istress=istress)
                stresses.append(stress)
        else:
            stress = ml.get_stress(name)
            if isinstance(stress, list):
                stresses.extend(stress)
            else:
                stresses.append(stress)
    return stresses


def share_xaxes(axes: List[Axes]) -> None:
    """share x-axes"""
    for i, iax in enumerate(axes):
        if i < (len(axes) - 1):
            iax.sharex(axes[-1])
            for t in iax.get_xticklabels():
                t.set_visible(False)


def share_yaxes(axes: List[Axes]) -> None:
    """share y-axes"""
    for iax in axes[1:]:
        iax.sharey(axes[0])
        for t in iax.get_yticklabels():
            t.set_visible(False)



"""The following methods may be used to calculate the crosscorrelation and
autocorrelation for a time series.

These methods are 'special' in the sense that they are able to deal with irregular
time steps often observed in hydrological time series.
"""

# Type Hinting
from typing import Tuple, Union

from numpy import (
    append,
    arange,
    array,
    average,
    corrcoef,
    diff,
    empty_like,
    exp,
    inf,
    nan,
    ones,
    pi,
    sqrt,
)
from pandas import DataFrame, Index, Series, Timedelta, to_timedelta
from scipy.stats import norm

from pastas.typing import ArrayLike

from ..decorators import njit


def acf(
    x: Series,
    lags: ArrayLike = 365,
    bin_method: str = "rectangle",
    bin_width: float = 0.5,
    max_gap: float = inf,
    min_obs: int = 20,
    full_output: bool = False,
    alpha: float = 0.05,
) -> Union[Series, DataFrame]:
    """Calculate the autocorrelation function for irregular time steps.

    Parameters
    ----------
    x: pandas.Series
        Pandas Series containing the values to calculate the cross-correlation on.
        The index has to be a Pandas.DatetimeIndex.
    lags: array_like, optional
        numpy array containing the lags in days for which the cross-correlation if
        calculated. Defaults is all lags from 1 to 365 days.
    bin_method: str, optional
        method to determine the type of bin. Options are "rectangle" (default),
        "gaussian" and "regular" (for regular timesteps).
    bin_width: float, optional
        number of days used as the width for the bin to calculate the correlation.
    max_gap: float, optional
        Maximum time step gap in the data. All time steps above this gap value are
        not used for calculating the average time step. This can be helpful when
        there is a large gap in the data that influences the average time step.
    min_obs: int, optional
        Minimum number of observations in a bin to determine the correlation.
    full_output: bool, optional
        If True, also estimated uncertainties are returned. Default is False.
    alpha: float, optional
        alpha level to compute the confidence interval (e.g., 1-alpha).

    Returns
    -------
    result: pandas.Series or pandas.DataFrame
        If full_output=True, a DataFrame with columns "acf", "conf", and "n",
        containning the autocorrelation function, confidence intervals (depends on
        alpha), and the number of samples n used to compute these, respectively. If
        full_output=False, only the ACF is returned.

    Notes
    -----
    Calculate the autocorrelation function for irregular timesteps based on the
    slotting technique. Different methods (kernels) to bin the data are available.
    Method here is based on :cite:t:`rehfeld_comparison_2011`.

    If the time series have regular time step we recommend to use the acf method from
    the Statsmodels package.

    Examples
    --------
    For example, to estimate the autocorrelation for every second lag up to lags of
    one year:

    >>> acf = ps.stats.acf(x, lags=np.arange(1.0, 366.0, 2.0))

    See Also
    --------
    pastas.stats.ccf
    statsmodels.api.tsa.acf
    """
    c = ccf(
        x=x,
        y=x,
        lags=lags,
        bin_method=bin_method,
        bin_width=bin_width,
        max_gap=max_gap,
        min_obs=min_obs,
        full_output=full_output,
        alpha=alpha,
    )
    c.name = "ACF"
    if full_output:
        return c.rename(columns={"ccf": "acf"})
    else:
        return c


def ccf(
    x: Series,
    y: Series,
    lags: ArrayLike = 365,
    bin_method: str = "rectangle",
    bin_width: float = 0.5,
    max_gap: float = inf,
    min_obs: int = 20,
    full_output: bool = False,
    alpha: float = 0.05,
) -> Union[Series, DataFrame]:
    """Method to compute the cross-correlation for irregular time series.

    Parameters
    ----------
    x,y: pandas.Series
        Pandas Series containing the values to calculate the cross-correlation on.
        The index has to be a Pandas.DatetimeIndex.
    lags: array_like, optional
        numpy array containing the lags in days for which the cross-correlation is
        calculated. Defaults is all lags from 1 to 365 days.
    bin_method: str, optional
        method to determine the type of bin. Options are "rectangle" (default),
        "gaussian" and "regular" (for regular timesteps).
    bin_width: float, optional
        number of days used as the width for the bin to calculate the correlation.
    max_gap: float, optional
        Maximum timestep gap in the data. All timesteps above this gap value are not
        used for calculating the average timestep. This can be helpful when there is
        a large gap in the data that influences the average timestep.
    min_obs: int, optional
        Minimum number of observations in a bin to determine the correlation.
    full_output: bool, optional
        If True, also estimated uncertainties are returned. Default is False.
    alpha: float
        alpha level to compute the confidence interval (e.g., 1-alpha).

    Returns
    -------
    result: pandas.Series or pandas.DataFrame
        If full_output=True, a DataFrame with columns "ccf", "conf", and "n",
        containning the cross-correlation function, confidence intervals (depends on
        alpha), and the number of samples n used to compute these, respectively. If
        full_output=False, only the CCF is returned.

    Examples
    --------
    >>> ccf = ps.stats.ccf(x, y, bin_method="gaussian")
    """
    # prepare the time indices for x and y
    if x.index.inferred_freq and y.index.inferred_freq:
        bin_method = "regular"
    elif bin_method == "regular":
        raise Warning(
            "time series does not have regular time steps, choose different bin_method."
        )

    x, t_x, dt_x_mu = _preprocess(x, max_gap=max_gap)
    y, t_y, dt_y_mu = _preprocess(y, max_gap=max_gap)
    dt_mu = max(dt_x_mu, dt_y_mu)  # The mean time step from both series

    if isinstance(lags, int) and bin_method == "regular":
        lags = arange(int(dt_mu), lags + 1, int(dt_mu), dtype=float)
    elif isinstance(lags, int):
        lags = arange(1.0, lags + 1, dtype=float)
    elif isinstance(lags, list):
        lags = array(lags, dtype=float)

    if bin_method == "rectangle":
        c, b = _compute_ccf_rectangle(lags, t_x, x, t_y, y, bin_width)
    elif bin_method == "gaussian":
        c, b = _compute_ccf_gaussian(lags, t_x, x, t_y, y, bin_width)
    elif bin_method == "regular":
        c, b = _compute_ccf_regular(arange(1.0, len(lags) + 1), x, y)
    else:
        raise NotImplementedError

    conf = norm.ppf(1 - alpha / 2.0) / sqrt(b)
    result = DataFrame(
        data={"ccf": c, "conf": conf, "n": b},
        dtype=float,
        index=Index(to_timedelta(lags, unit="D"), name="Lags"),
    )

    result = result.where(result.n > min_obs).dropna()

    if full_output:
        return result
    else:
        return result.ccf


def _preprocess(x: Series, max_gap: float) -> Tuple[ArrayLike, ArrayLike, float]:
    """Internal method to preprocess the time series."""
    dt = x.index.to_series().diff().dropna().values / Timedelta(1, "D")
    dt_mu = dt[dt < max_gap].mean()  # Deal with big gaps if present
    if int(dt_mu) == 0:
        dt_mu = 1  # Prevent division by zero error
    t = dt.cumsum()

    # Normalize the values and create numpy arrays
    x = (x.values - x.values.mean()) / x.values.std()

    return x, t, dt_mu


@njit
def _compute_ccf_rectangle(
    lags: ArrayLike,
    t_x: ArrayLike,
    x: ArrayLike,
    t_y: ArrayLike,
    y: ArrayLike,
    bin_width: float = 0.5,
) -> Tuple[ArrayLike, ArrayLike]:
    """Internal numba-optimized method to compute the ccf."""
    c = empty_like(lags)
    b = empty_like(lags)
    n = len(t_x)

    for k in range(len(lags)):
        cl = 0.0
        b_sum = 0.0
        for i in range(n):
            for j in range(n):
                d = abs(t_x[i] - t_y[j]) - lags[k]
                if abs(d) <= bin_width:
                    cl += x[i] * y[j]
                    b_sum += 1
        if b_sum == 0.0:
            c[k] = nan
            b[k] = 0.01  # Prevent division by zero error
        else:
            c[k] = cl / b_sum
            b[k] = b_sum / 2  # divide by 2 because we over count in for-loop
    return c, b


@njit
def _compute_ccf_gaussian(
    lags: ArrayLike,
    t_x: ArrayLike,
    x: ArrayLike,
    t_y: ArrayLike,
    y: ArrayLike,
    bin_width: float = 0.5,
) -> Tuple[ArrayLike, ArrayLike]:
    """Internal numba-optimized method to compute the ccf."""
    c = empty_like(lags)
    b = empty_like(lags)
    n = len(t_x)

    den1 = -2 * bin_width**2  # denominator 1
    den2 = sqrt(2 * pi * bin_width)  # denominator 2

    for k in range(len(lags)):
        cl = 0.0
        b_sum = 0.0

        for i in range(n):
            for j in range(n):
                d = t_x[i] - t_y[j] - lags[k]
                d = exp(d**2 / den1) / den2
                cl += x[i] * y[j] * d
                b_sum += d
        if b_sum == 0.0:
            c[k] = nan
            b[k] = 0.01  # Prevent division by zero error
        else:
            c[k] = cl / b_sum
            b[k] = b_sum / 2  # divide by 2 because we over count in for-loop
    return c, b


def _compute_ccf_regular(
    lags: ArrayLike, x: ArrayLike, y: ArrayLike
) -> Tuple[ArrayLike, ArrayLike]:
    c = empty_like(lags)
    for i, lag in enumerate(lags):
        c[i] = corrcoef(x[: -int(lag)], y[int(lag) :])[0, 1]
    b = len(x) - lags
    return c, b


def mean(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) mean of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    Notes
    -----
    The (weighted) mean for a time series x is computed as:

    .. math:: \\bar{x} = \\sum_{i=1}^{N} w_i x_i

    where :math:`w_i` are the weights, taken as the time step between observations,
    normalized by the sum of all time steps.
    """
    w = _get_weights(x, weighted=weighted, max_gap=max_gap)
    return average(x.to_numpy(), weights=w)


def var(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) variance of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    Notes
    -----
    The (weighted) variance for a time series x is computed as:

    .. math:: \\sigma_x^2 = \\sum_{i=1}^{N} w_i (x_i - \\bar{x})^2

    where :math:`w_i` are the weights, taken as the time step between observations,
    normalized by the sum of all time steps. Note how weighted mean (:math:`\\bar{
    x}`) is used in this formula.
    """
    w = _get_weights(x, weighted=weighted, max_gap=max_gap)
    mu = average(x.to_numpy(), weights=w)
    sigma = (x.size / (x.size - 1) * w * (x.to_numpy() - mu) ** 2).sum()
    return sigma


def std(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) variance of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    See Also
    --------
    ps.stats.mean, ps.stats.var
    """
    return sqrt(var(x, weighted=weighted, max_gap=max_gap))


# Helper functions


def _get_weights(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Helper method to compute the weights as the time step between obs.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 30 days.
    """
    if weighted:
        w = append(0.0, diff(x.index.to_numpy()) / Timedelta("1D"))
        w[w > max_gap] = max_gap
    else:
        w = ones(x.index.size)
    w /= w.sum()
    return w



from pandas import Series

from pastas import Model
from pastas.plotting.plots import TrackSolve, compare, pairplot

# mpl.use("Agg")  # prevent _tkinter.TclError: Can't find a usable tk.tcl error


def test_plot(ml: Model) -> None:
    _ = ml.plot()


def test_decomposition(ml: Model) -> None:
    _ = ml.plots.decomposition(min_ylim_diff=0.1)


def test_results(ml: Model) -> None:
    _ = ml.plots.results()


def test_stacked_results(ml: Model) -> None:
    _ = ml.plots.stacked_results()


def test_block_response(ml: Model) -> None:
    _ = ml.plots.block_response()


def test_step_response(ml: Model) -> None:
    _ = ml.plots.step_response()


def test_diagnostics(ml: Model) -> None:
    _ = ml.plots.diagnostics(acf_options=dict(min_obs=10))


def test_stresses(ml: Model) -> None:
    _ = ml.plots.stresses()


def test_contributions_pie(ml: Model) -> None:
    _ = ml.plots.contributions_pie()


def test_compare(ml: Model) -> None:
    ml2 = ml.copy()
    models = [ml, ml2]
    _ = compare(models, names=["ml1", "ml2"])


def test_tracksolve(ml: Model) -> None:
    track = TrackSolve(ml)
    _ = ml.solve(callback=track.plot_track_solve)


def test_summary_pdf(ml: Model) -> None:
    _ = ml.plots.summary_pdf()


def test_pairplot(prec: Series, pevap: Series, head: Series) -> None:
    _ = pairplot([prec, pevap, head])



# flake8: noqa
# Type hinting for Pastas library
# Typing
from typing import TYPE_CHECKING, Any, Callable, TypedDict, TypeVar, Union

# External libraries
# Matplotlib
from matplotlib.axes import Axes as MatplotlibAxes
from matplotlib.figure import Figure as MatplotlibFigure

# Numpy
from numpy.typing import ArrayLike as NumpyArrayLike

# Pandas
from pandas import Timestamp

# External Types
Axes = TypeVar("Axes", bound=MatplotlibAxes)  # Matplotlib Axes
Figure = TypeVar("Figure", bound=MatplotlibFigure)  # Matplotlib Figure
ArrayLike = TypeVar("ArrayLike", bound=NumpyArrayLike)  # Array Like (NumPy based)

# Internal library
if TYPE_CHECKING:  # https://mypy.readthedocs.io/en/latest/runtime_troubles.html
    import pastas as ps

# Internal Types
TimestampType = TypeVar("TimestampType", bound=Union[str, Timestamp])  # Tmin or Tmax
Model = TypeVar("Model", bound="ps.Model")  # Model
TimeSeries = TypeVar("TimeSeries", bound="ps.timeseries.TimeSeries")  # Time Series
StressModel = TypeVar(
    "StressModel", bound="ps.stressmodels.StressModelBase"
)  # Stress Model
NoiseModel = TypeVar("NoiseModel", bound="ps.noisemodels.NoiseModelBase")  # Noise Model
Solver = TypeVar("Solver", bound="ps.solver.BaseSolver")  # Base Solver
Recharge = TypeVar("Recharge", bound="ps.recharge.RechargeBase")  # Recharge Base
CallBack = TypeVar("CallBack", bound=Any)  # Callback
Function = Callable[..., Any]  # Function (e.g. Objective Function)
RFunc = TypeVar("RFunc", bound="ps.rfunc.RfuncBase")  # rFunc Base


class OseriesSettingsDict(TypedDict):
    """
    Time series settings is a dictionary defining logic for filling and up- or
    downsampling time series.

    Time series settings
    --------------------
    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    sample_down : {"drop", "mean", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `drop`: resample the time series by taking the mean, dropping any NaN-values
           * `mean`: resample time series by taking the mean
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value
    """

    sample_down: str
    fill_nan: str


class StressSettingsDict(TypedDict):
    """
    Time series settings is a dictionary defining logic for filling and up- or
    downsampling time series.

    Time series settings
    --------------------
    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    fill_before : {"mean", "bfill"} or float
        Method for extending time series into past.
           * `mean`: extend time series into past with mean value of time series
           * `bfill`: extend time series into past by back-filling first value
           * `float`: extend time series into past with provided value, e.g. 0.0
    fill_after : {"mean", "ffill"} or float
        Method for extending time series into future.
           * `mean`: extend time series into future with mean value of time series
           * `ffill`: extend time series into future by forward-filling last value
           * `float`: extend time series into future with provided value, e.g. 0.0
    sample_up : {"mean", "interpolate", "divide"} or float
        Method for up-sampling time series (increasing frequency, e.g. going from weekly
        to daily values).
           * `bfill` or `backfill`: fill up-sampled time steps by back-filling current
             values
           * `ffill` or `pad`: fill up-sampled time steps by forward-filling current
             values
           * `mean`: fill up-sampled time steps with mean of timeseries
           * `interpolate`: fill up-sampled time steps by interpolating between current
             values
           * `divide`: fill up-sampled steps with current value divided by length of
             current time steps (i.e. spread value over new time steps).
    sample_down : {"mean", "drop", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `mean`: resample time series by taking the mean
           * `drop`: resample the time series by taking the mean, dropping any
             NaN-values
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value
    """

    sample_up: str
    sample_down: str
    fill_nan: Union[str, float]
    fill_before: Union[str, float]
    fill_after: Union[str, float]



# flake8: noqa
import logging
import warnings

from pandas.plotting import register_matplotlib_converters

import pastas.objective_functions as objfunc
import pastas.plotting.plots as plots
import pastas.recharge as rch
import pastas.stats as stats
import pastas.timeseries_utils as ts
from pastas import extensions
from pastas.dataset import list_datasets, load_dataset
from pastas.decorators import set_use_numba
from pastas.model import Model
from pastas.noisemodels import ArmaModel, ArmaNoiseModel, ArNoiseModel, NoiseModel
from pastas.plotting.modelcompare import CompareModels
from pastas.plotting.plots import TrackSolve
from pastas.rcparams import rcParams
from pastas.rfunc import (
    DoubleExponential,
    Exponential,
    FourParam,
    Gamma,
    Hantush,
    HantushWellModel,
    Kraijenhoff,
    One,
    Polder,
    Spline,
)
from pastas.solver import EmceeSolve, LeastSquares, LmfitSolve
from pastas.stressmodels import (
    ChangeModel,
    Constant,
    LinearTrend,
    RechargeModel,
    StepModel,
    StressModel,
    TarsoModel,
    WellModel,
)
from pastas.timeseries import validate_oseries, validate_stress
from pastas.transform import ThresholdTransform
from pastas.utils import set_log_level
from pastas.version import __version__, show_versions


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg)


warnings.formatwarning = custom_formatwarning
warnings.warn(
    """DeprecationWarning: As of Pastas 1.5, no noisemodel is added to the pastas Model class by default anymore. To solve your model using a noisemodel, you have to explicitly add a noisemodel to your model before solving. For more information, and how to adapt your code, please see this issue on GitHub: https://github.com/pastas/pastas/issues/735""",
    DeprecationWarning,
    stacklevel=2,
)

logger = logging.getLogger(__name__)

# Register matplotlib converters when using Pastas
# https://github.com/pastas/pastas/issues/92

register_matplotlib_converters()



"""This module contains utility functions for working with Pastas models."""

import logging
from logging import handlers
from platform import platform

# Type Hinting
from typing import Any, Optional, Tuple

from pandas import Timestamp

from pastas.typing import Model as ModelType
from pastas.typing import TimestampType

logger = logging.getLogger(__name__)


def get_stress_tmin_tmax(ml: ModelType) -> Tuple[TimestampType, TimestampType]:
    """Get the minimum and maximum time that all the stresses have data."""
    from pastas import Model

    tmin = Timestamp.min
    tmax = Timestamp.max
    if isinstance(ml, Model):
        for sm in ml.stressmodels:
            for st in ml.stressmodels[sm].stress:
                tmin = max((tmin, st.series_original.index.min()))
                tmax = min((tmax, st.series_original.index.max()))
    else:
        raise (TypeError("Unknown type {}".format(type(ml))))
    return tmin, tmax


def initialize_logger(
    logger: Optional[Any] = None, level: Optional[Any] = logging.INFO
) -> None:
    """Internal method to create a logger instance to log program output.

    Parameters
    ----------
    logger : logging.Logger
        A Logger-instance. Use pastas.logger to initialise the Logging instance that
        handles all logging throughout pastas, including all submodules and packages.
    """
    if logger is None:
        logger = logging.getLogger("pastas")
    logger.setLevel(level)
    remove_file_handlers(logger)
    set_console_handler(logger)
    # add_file_handlers(logger)


def set_console_handler(
    logger: Optional[Any] = None,
    level: Optional[Any] = logging.INFO,
    fmt: str = "%(levelname)s: %(message)s",
) -> None:
    """Method to add a console handler to the logger of Pastas.

    Parameters
    ----------
    logger : logging.Logger
        A Logger-instance. Use ps.logger to initialise the Logging instance that
        handles all logging throughout pastas,  including all submodules and packages.
    """
    if logger is None:
        logger = logging.getLogger("pastas")
    remove_console_handler(logger)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter(fmt=fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def set_log_level(level: str) -> None:
    """Set the log-level for Pastas.

    Parameters
    ----------
    level: str
        String with the level to log messages to the screen for. Options are: "INFO",
        "WARNING", and "ERROR".

    Examples
    --------

    >>> import pandas as ps
    >>> ps.set_log_level("ERROR")
    """
    logger = logging.getLogger("pastas")
    logger.setLevel(level)


def remove_console_handler(logger: Optional[Any] = None) -> None:
    """Method to remove the console handler to the logger of Pastas.

    Parameters
    ----------
    logger : logging.Logger
        A Logger-instance. Use ps.logger to initialise the Logging instance that
        handles all logging throughout pastas, including all submodules and packages.
    """
    if logger is None:
        logger = logging.getLogger("pastas")
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            logger.removeHandler(handler)


def add_file_handlers(
    logger: Optional[Any] = None,
    filenames: Tuple[str] = ("info.log", "errors.log"),
    levels: Tuple[Any] = (logging.INFO, logging.ERROR),
    maxBytes: int = 10485760,
    backupCount: int = 20,
    encoding: str = "utf8",
    fmt: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt: str = "%y-%m-%d %H:%M",
) -> None:
    """Method to add file handlers in the logger of Pastas.

    Parameters
    ----------
    logger : logging.Logger
        A Logger-instance. Use ps.logger to initialise the Logging instance that
        handles all logging throughout pastas, including all submodules and packages.
    """
    if logger is None:
        logger = logging.getLogger("pastas")
    # create formatter
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)

    # create file handlers, set the level & formatter, and add it to the logger
    for filename, level in zip(filenames, levels):
        fh = handlers.RotatingFileHandler(
            filename, maxBytes=maxBytes, backupCount=backupCount, encoding=encoding
        )
        fh.setLevel(level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)


def remove_file_handlers(logger: Optional[logging.Logger] = None) -> None:
    """Method to remove any file handlers in the logger of Pastas.

    Parameters
    ----------
    logger : logging.Logger
        A Logger-instance. Use ps.logger to initialise the Logging instance that
        handles all logging throughout pastas, including all submodules and packages.
    """
    if logger is None:
        logger = logging.getLogger("pastas")
    for handler in logger.handlers:
        if isinstance(handler, handlers.RotatingFileHandler):
            logger.removeHandler(handler)


def validate_name(name: str, raise_error: bool = False) -> str:
    """Method to check user-provided names and log a warning if wrong.

    Parameters
    ----------
    name: str
        String with the name to check for illegal characters.
    raise_error: bool
        raise Exception error if illegal character is found, default is False which
        only logs a warning.

    Returns
    -------
    name: str
        Unchanged name string.
    """
    ilchar = ["/", "\\", " ", ".", "'", '"', "`"]
    if "windows" in platform().lower():
        ilchar += [
            "#",
            "%",
            "&",
            "@",
            "{",
            "}",
            "|",
            "$",
            "*",
            "<",
            ">",
            "?",
            "!",
            ":",
            "=",
            "+",
        ]

    name = str(name)
    for char in ilchar:
        if char in name:
            msg = (
                "User-provided name '%s' contains illegal character. Please "
                "remove '%s' from name."
            )
            if raise_error:
                logger.error(msg, name, char)
                raise Exception(msg % (name, char))
            else:
                logger.warning(msg, name, char)

    return name



import numpy as np
import pandas as pd

import pastas as ps


def acf_func(**kwargs):
    index = pd.to_datetime(np.arange(0, 100, 1), unit="D", origin="2000")
    index = kwargs.pop("index", index)
    data = np.sin(np.linspace(0, 10 * np.pi, 100))
    r = pd.Series(data=data, index=index)
    lags = np.arange(1.0, 11.0)
    lags = kwargs.pop("lags", lags)
    acf_true_len = lags + 1 if isinstance(lags, int) else len(lags) + 1
    acf_true = np.cos(np.linspace(0.0, np.pi, acf_true_len))[1:]
    acf = ps.stats.acf(r, lags=lags, min_obs=1, **kwargs).values
    return acf, acf_true


def test_acf_rectangle() -> None:
    acf, acf_true = acf_func(bin_method="rectangle")
    assert abs((acf - acf_true)).max() < 0.05


def test_acf_gaussian():
    acf, acf_true = acf_func(bin_method="gaussian")
    assert abs((acf - acf_true)).max() < 0.05


def test_acf_hourly():
    index = pd.to_datetime(np.arange(0, 100, 1), unit="h", origin="2000")
    lags = 10
    acf, acf_true = acf_func(index=index, lags=lags)
    assert abs((acf - acf_true)).max() < 0.05


def test_runs_test():
    """
    http://www.itl.nist.gov/div898/handbook/eda/section3/eda35d.htm
    True Z-statistic = 2.69
    Read NIST test data
    """
    data = pd.read_csv("tests/data/nist.csv")
    test, _ = ps.stats.runs_test(data)
    assert test[0] - 2.69 < 0.02


def test_stoffer_toloi():
    res = pd.Series(
        index=pd.date_range(start=0, periods=1000, freq="D"), data=np.random.rand(1000)
    )
    _, pval = ps.stats.stoffer_toloi(res)
    assert pval > 1e-10



from numpy import arange, isclose, sin
from pandas import Series, read_csv

import pastas as ps

# Load series before
rain = (
    read_csv("tests/data/rain.csv", index_col=0, parse_dates=True)
    .squeeze("columns")
    .loc["2005":]
    * 1e3
)
evap = (
    read_csv("tests/data/evap.csv", index_col=0, parse_dates=True)
    .squeeze("columns")
    .loc["2005":]
    * 1e3
)
temp = Series(index=evap.index, data=sin(arange(evap.size) / 365 * 6), dtype=float)


def test_model_solve(ml: ps.Model) -> None:
    ml.solve()


def test_model_copy(ml: ps.Model) -> None:
    ml.copy()


def test_berendrecht(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(prec=rain, evap=evap, recharge=ps.rch.Berendrecht())
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_peterson(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(prec=rain, evap=evap, recharge=ps.rch.Peterson())
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_linear(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(prec=rain, evap=evap, recharge=ps.rch.Linear())
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_flexmodel(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(prec=rain, evap=evap, recharge=ps.rch.FlexModel())
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_flexmodel_no_interception(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(
        prec=rain, evap=evap, recharge=ps.rch.FlexModel(interception=False)
    )
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_flexmodel_gw_uptake(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(
        prec=rain, evap=evap, recharge=ps.rch.FlexModel(gw_uptake=True)
    )
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_flexmodel_snow(ml_empty: ps.Model) -> None:
    rm = ps.RechargeModel(
        prec=rain, evap=evap, temp=temp, recharge=ps.rch.FlexModel(snow=True)
    )
    ml_empty.add_stressmodel(rm)
    ml_empty.solve()


def test_flexmodel_water_balance_rootzone() -> None:
    rch = ps.rch.FlexModel()
    e = evap.to_numpy()
    p = rain.to_numpy()
    sr, r, ea, q, pe = rch.get_root_zone_balance(p, e)
    error = sr[0] - sr[-1] + (r + ea + q + pe)[:-1].sum()
    assert isclose(error, 0)


def test_flexmodel_water_balance_snow() -> None:
    rch = ps.rch.FlexModel()
    p = rain.to_numpy()
    t = temp.to_numpy()
    ss, snow, m = rch.get_snow_balance(p, t)
    error = ss[0] - ss[-1] + (snow + m)[:-1].sum()
    assert isclose(error, 0)


def test_flexmodel_water_balance_interception() -> None:
    rch = ps.rch.FlexModel()
    e = evap.to_numpy()
    p = rain.to_numpy()
    si, ei, pi = rch.get_interception_balance(p, e)
    error = si[0] - si[-1] + (pi + ei)[:-1].sum()
    assert isclose(error, 0)



import numpy as np
import pandas as pd

import pastas as ps


class TestQGXG(object):
    def test_q_ghg(self) -> None:
        n = 101
        idx = pd.date_range("20160101", freq="d", periods=n)
        s = pd.Series(np.arange(n), index=idx)
        v = ps.stats.q_ghg(s, q=0.94)
        assert v == 94.0

    def test_q_glg(self) -> None:
        n = 101
        idx = pd.date_range("20160101", freq="d", periods=n)
        s = pd.Series(np.arange(n), index=idx)
        v = ps.stats.q_glg(s, q=0.06)
        assert v == 6.0

    def test_q_ghg_nan(self) -> None:
        idx = pd.date_range("20160101", freq="d", periods=4)
        s = pd.Series([1, np.nan, 3, np.nan], index=idx)
        v = ps.stats.q_ghg(s, q=0.5)
        assert v == 2.0

    def test_q_gvg(self) -> None:
        idx = pd.to_datetime(["20160320", "20160401", "20160420"])
        s = pd.Series([0, 5, 10], index=idx)
        v = ps.stats.q_gvg(s)
        assert v == 2.5

    def test_q_gvg_nan(self) -> None:
        idx = pd.to_datetime(["20160820", "20160901", "20161120"])
        s = pd.Series([0, 5, 10], index=idx)
        v = ps.stats.q_gvg(s)
        assert np.isnan(v)

    def test_q_glg_tmin(self) -> None:
        tmin = "20160301"
        idx = pd.date_range("20160101", "20160331", freq="d")
        s = pd.Series(np.arange(len(idx)), index=idx)
        v = ps.stats.q_glg(s, q=0.06, tmin=tmin)
        assert v == 61.8

    def test_q_ghg_tmax(self) -> None:
        n = 101
        tmax = "20160301"
        idx = pd.date_range("20160101", freq="d", periods=n)
        s = pd.Series(np.arange(n), index=idx)
        v = ps.stats.q_ghg(s, q=0.94, tmax=tmax)
        assert v == 56.4

    def test_q_gvg_tmin_tmax(self) -> None:
        tmin = "20170301"
        tmax = "20170401"
        idx = pd.to_datetime(["20160401", "20170401", "20180401"])
        s = pd.Series([0, 5, 10], index=idx)
        v = ps.stats.q_gvg(s, tmin=tmin, tmax=tmax)
        assert v == 5

    def test_q_gxg_series(self) -> None:
        s = pd.read_csv(
            "tests/data/hseries_gxg.csv",
            index_col=0,
            header=0,
            parse_dates=True,
            dayfirst=True,
        ).squeeze("columns")
        ghg = ps.stats.q_ghg(s)
        glg = ps.stats.q_glg(s)
        gvg = ps.stats.q_gvg(s)
        print("\n")
        print("calculated GXG's percentile method: \n")
        print(
            (
                "GHG: {ghg:.2f} m+NAP\n"
                "GLG: {glg:.2f} m+NAP\n"
                "GVG: {gvg:.2f} m+NAP\n"
            ).format(ghg=ghg, glg=glg, gvg=gvg)
        )
        print("Menyanthes GXG's: \n")
        print(
            (
                "GHG: {ghg:.2f} m+NAP\n"
                "GLG: {glg:.2f} m+NAP\n"
                "GVG: {gvg:.2f} m+NAP\n"
            ).format(ghg=-3.23, glg=-3.82, gvg=-3.43)
        )



import pastas as ps


def test_least_squares(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())


def test_fit_constant(ml: ps.Model):
    ml.solve(fit_constant=False)


def test_no_noise(ml: ps.Model):
    ml.del_noisemodel()
    ml.solve()


# test the uncertainty method here
def test_pred_interval(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.solver.prediction_interval(n=10)


def test_ci_simulation(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.solver.ci_simulation(n=10)


def test_ci_block_response(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.solver.ci_block_response(name="rch", n=10)


def test_ci_step_response(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.solver.ci_step_response(name="rch", n=10)


def test_ci_contribution(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.solver.ci_contribution(name="rch", n=10)


# Test the EmceeSolver


def test_emcee(ml: ps.Model):
    ml.solve(solver=ps.LeastSquares())
    ml.del_noisemodel()
    ml.solve(
        solver=ps.EmceeSolve(nwalkers=20),
        initial=False,
        fit_constant=False,
        steps=10,
    )



import os

import matplotlib.pyplot as plt
import pytest

pathname = os.path.join("doc", "examples")
# get list of examples to run
files = [f for f in os.listdir(pathname) if f.endswith(".py")]


@pytest.mark.parametrize("file", files)
def test_example(file) -> None:
    cwd = os.getcwd()
    os.chdir(pathname)
    try:
        # run each example
        exec(open(file).read())
        plt.close("all")
    except Exception as e:
        os.chdir(cwd)
        raise Exception(f"could not run {file}") from e
    os.chdir(cwd)


if __name__ == "__main__":
    for file in files:
        test_example(file)



import numpy as np
import pytest
from pandas import Series, Timedelta, date_range, read_csv

import pastas as ps

obs = (
    read_csv("tests/data/obs.csv", index_col=0, parse_dates=True)
    .squeeze("columns")
    .dropna()
)
prec = read_csv("tests/data/rain.csv", index_col=[0], parse_dates=True).squeeze() * 1e-3
evap = read_csv("tests/data/evap.csv", index_col=[0], parse_dates=True).squeeze() * 1e-3


def generate_synthetic_heads(input, rfunc, params, const=10.0, cutoff=0.999, dt=1.0):
    # Generate the head
    step = rfunc.block(params, cutoff=cutoff, dt=dt)

    h = const * np.ones(len(input) + step.size)

    for i in range(len(input)):
        h[i : i + step.size] += input.iloc[i] * step

    head = Series(
        index=input.index,
        data=h[: len(input)],
    )
    return head


def test_create_model() -> None:
    ml = ps.Model(obs, name="Test_Model")
    ml.add_noisemodel(ps.ArNoiseModel())


def test_add_stressmodel(ml_empty: ps.Model, sm_prec) -> None:
    ml_empty.add_stressmodel(sm_prec)


def test_add_stressmodels(
    ml_empty: ps.Model, sm_prec: ps.StressModel, sm_evap: ps.StressModel
) -> None:
    ml_empty.add_stressmodel([sm_prec, sm_evap])


def test_del_stressmodel(ml: ps.Model) -> None:
    ml.del_stressmodel("rch")


def test_add_constant(ml_empty: ps.Model) -> None:
    ml_empty.add_constant(ps.Constant())


def test_del_constant(ml_empty: ps.Model) -> None:
    ml_empty.del_constant()


def test_add_noisemodel(ml_empty: ps.Model) -> None:
    ml_empty.add_noisemodel(ps.ArNoiseModel())


def test_armamodel() -> None:
    ml = ps.Model(obs, name="Test_Model")
    ml.add_noisemodel(ps.ArmaNoiseModel())
    ml.solve()
    ml.to_file("test.pas")


def test_del_noisemodel(ml_empty: ps.Model) -> None:
    ml_empty.del_noisemodel()


def test_solve_model(ml: ps.Model) -> None:
    ml.solve()


def test_solve_empty_model(ml: ps.Model) -> None:
    with pytest.raises(ValueError) as excinfo:
        ml.solve(tmin="2016")
    assert excinfo.value.args[0].startswith("Calibration series")


def test_save_model(ml: ps.Model) -> None:
    ml.to_file("test.pas")


def test_load_model(ml: ps.Model) -> None:
    ml.solve()
    # add some fictitious tiny value for testing float precision
    ml.parameters.loc["rch_f", "pmax"] = 1.23456789e-10
    ml.to_file("test.pas")
    ml2 = ps.io.load("test.pas")

    # dataframe dtypes don't match... make the same here
    # this is caused because the parameters df is loaded empty without
    # information on the datatype in each column
    for icol in ["initial", "optimal", "pmin", "pmax", "stderr"]:
        ml.parameters[icol] = ml.parameters[icol].astype(float)
    ml.parameters["vary"] = ml.parameters["vary"].astype(bool)

    # check if parameters and pcov dataframes are equal
    assert ml.parameters.equals(ml2.parameters)
    assert ml.solver.pcov.equals(ml2.solver.pcov)


def test_model_copy(ml_empty: ps.Model) -> None:
    ml_empty.copy()


def test_get_block(ml: ps.Model) -> None:
    ml.get_block_response("rch")


def test_get_step(ml: ps.Model) -> None:
    ml.get_step_response("rch")


def test_get_contribution(ml: ps.Model) -> None:
    ml.get_contribution("rch")


def test_get_stress(ml: ps.Model) -> None:
    ml.get_stress("rch")


def test_simulate(ml: ps.Model) -> None:
    ml.simulate()


def test_residuals(ml: ps.Model) -> None:
    ml.residuals()


def test_noise(ml: ps.Model) -> None:
    ml.noise()


def test_observations(ml_empty: ps.Model) -> None:
    ml_empty.observations()


def test_get_output_series(ml: ps.Model) -> None:
    ml.get_output_series()


def test_get_output_series_arguments(ml: ps.Model) -> None:
    ml.get_output_series(split=False, add_contributions=False)


def test_model_sim_w_nans_error(ml_no_settings):
    with pytest.raises(ValueError) as _:
        ml_no_settings.solve()


def test_modelstats(ml: ps.Model) -> None:
    ml.solve()
    ml.stats.summary()


def test_fit_report(ml: ps.Model) -> None:
    ml.solve(report=False)
    ml.fit_report(corr=True, stderr=True)


def test_model_freq_geq_daily() -> None:
    rf_rch = ps.Exponential()
    A_rch = 800
    a_rch = 50
    f_rch = -1.3
    constant = 20

    stress = prec + f_rch * evap
    head = generate_synthetic_heads(stress, rf_rch, (A_rch, a_rch), const=constant)

    models = []
    freqs = ["1D", "7D", "14D", "28D"]
    for freq in freqs:
        iml = ps.Model(head, name=freq)
        rm = ps.RechargeModel(prec, evap, rfunc=rf_rch, name="recharge")
        iml.add_stressmodel(rm)
        iml.solve(freq=freq, report=False)
        models.append(iml)

    comparison = ps.CompareModels(models)
    assert (comparison.get_metrics(metric_selection=["rsq"]).squeeze() > 0.99).all()


def test_model_freq_h():
    rf_tide = ps.Exponential()
    A_tide = 1.0
    a_tide = 0.15

    # sine with period 12 hrs 25 minutes and amplitude 1.5 m
    tidx = date_range(obs.index[0], obs.index[-1] + Timedelta(hours=23), freq="h")
    tides = Series(
        index=tidx,
        data=1.5 * np.sin(2 * np.pi * np.arange(tidx.size) / (0.517375)),
    )

    ht = generate_synthetic_heads(tides, rf_tide, (A_tide, a_tide), dt=1 / 24.0)

    # model with hourly timestep
    ml_h = ps.Model(ht, name="tidal_model", freq="h")
    sm = ps.StressModel(
        tides,
        rfunc=ps.Exponential(),
        name="tide",
        settings="waterlevel",
    )
    ml_h.add_stressmodel(sm)
    ml_h.solve(report=False)

    assert ml_h.simulate().index.freq == "h"
    assert ml_h.stats.rsq() > 0.99999



import pandas as pd
import pytest

import pastas as ps

index = pd.date_range("2010-01-01", "2010-01-04", freq="D")
obs = pd.Series([10.0, 15.21, 20.0, 25.2], index=index)
sim = pd.Series([10.1, 15.0, 20.3, 25.0], index=index)
tol = 1e-6


def test_mae():
    mae = ps.stats.metrics.mae(obs=obs, sim=sim)
    assert pytest.approx(mae, tol) == 0.2025


def test_rmse():
    rmse = ps.stats.metrics.rmse(obs=obs, sim=sim)
    assert pytest.approx(rmse, tol) == 0.2145343


def test_sse():
    sse = ps.stats.metrics.sse(obs=obs, sim=sim)
    assert pytest.approx(sse, tol) == 0.1841


def test_pearsonr():
    r = ps.stats.metrics.pearsonr(obs=obs, sim=sim)
    assert pytest.approx(r, tol) == 0.999299


def test_evp():
    evp = ps.stats.metrics.evp(obs=obs, sim=sim)
    assert pytest.approx(evp, tol) == 99.85505


def test_nse():
    nse = ps.stats.metrics.nse(obs=obs, sim=sim)
    assert pytest.approx(nse, tol) == 0.99855


def test_rsq():
    rsq = ps.stats.metrics.rsq(obs=obs, sim=sim)
    assert pytest.approx(rsq, tol) == 0.99855


def test_bic():
    bic = ps.stats.metrics.bic(obs=obs, sim=sim)
    assert pytest.approx(bic, tol) == -10.9279878


def test_aic():
    aic = ps.stats.metrics.aic(obs=obs, sim=sim)
    assert pytest.approx(aic, tol) == -10.314282


def test_aicc():
    aicc = ps.stats.metrics.aicc(obs=obs, sim=sim)
    assert pytest.approx(aicc, tol) == -8.314282


def test_kge():
    kge = ps.stats.metrics.kge(obs=obs, sim=sim)
    assert pytest.approx(kge, tol) == 0.9923303


def test_kge_modified():
    kgem = ps.stats.metrics.kge(obs=obs, sim=sim, modified=True)
    assert pytest.approx(kgem, tol) == 0.99247



from functools import wraps
from logging import getLogger
from typing import Dict, Optional

from pastas.typing import Function, TimestampType

logger = getLogger(__name__)

USE_NUMBA = True


def set_use_numba(b: bool) -> None:
    global USE_NUMBA
    USE_NUMBA = b


def set_parameter(function: Function) -> Function:
    @wraps(function)
    def _set_parameter(self, name: str, value: float, **kwargs):
        if name not in self.parameters.index:
            msg = "Parameter name %s does not exist, please choose from %s"
            logger.error(msg, name, self.parameters.index)
            raise KeyError(msg % (name, self.parameters.index))
        else:
            return function(self, name, value, **kwargs)

    return _set_parameter


def get_stressmodel(function: Function) -> Function:
    @wraps(function)
    def _get_stressmodel(self, name: str, **kwargs):
        if name not in self.stressmodels.keys():
            msg = (
                "The stressmodel name you provided is not in the stressmodels dict. "
                "Please select from the following list: %s"
            )
            logger.error(msg, self.stressmodels.keys())
            raise KeyError(msg % self.stressmodels.keys())
        else:
            return function(self, name, **kwargs)

    return _get_stressmodel


def model_tmin_tmax(function: Function) -> Function:
    @wraps(function)
    def _model_tmin_tmax(
        self,
        tmin: Optional[TimestampType] = None,
        tmax: Optional[TimestampType] = None,
        *args,
        **kwargs,
    ):
        if tmin is None:
            tmin = self.ml.settings["tmin"]
        if tmax is None:
            tmax = self.ml.settings["tmax"]

        return function(self, tmin, tmax, *args, **kwargs)

    return _model_tmin_tmax


def PastasDeprecationWarning(function: Function) -> Function:
    @wraps(function)
    def _function(*args, **kwargs):
        logger.warning(
            "Method is deprecated since 1.5 and will be removed in Pastas version 1.6"
        )
        return function(*args, **kwargs)

    return _function


def njit(function: Optional[Function] = None, parallel: bool = False) -> Function:
    def njit_decorator(f: Function):
        try:
            if not USE_NUMBA:
                return f
            else:
                from numba import njit

                fnumba = njit(f, parallel=parallel)
                return fnumba
        except ImportError:
            return f

    if function:
        return njit_decorator(function)

    return njit_decorator


def latexfun(
    function: Optional[Function] = None,
    identifiers: Optional[Dict[str, str]] = None,
    use_math_symbols: bool = True,
    use_raw_function_name: bool = False,
) -> Function:
    def latexify_decorator(f: Function) -> Function:
        try:
            import latexify

            flatex = latexify.function(
                f,
                identifiers=identifiers,
                use_math_symbols=use_math_symbols,
                use_raw_function_name=use_raw_function_name,
            )
            return flatex
        except ImportError:
            return f

    if function:
        return latexify_decorator(function)

    return latexify_decorator



import pytest

import pastas as ps


@pytest.mark.parametrize("rfunc_name", ps.rfunc.__all__)
def test_rfunc(rfunc_name) -> None:
    if rfunc_name not in []:
        rfunc = getattr(ps.rfunc, rfunc_name)()
        if rfunc_name == "HantushWellModel":
            rfunc.set_distances(100.0)
        p = rfunc.get_init_parameters("test").initial.to_numpy()
        rfunc.block(p)
        rfunc.step(p)


@pytest.mark.parametrize("rfunc_name", ps.rfunc.__all__)
def test_to_dict_rfuncs(rfunc_name) -> None:
    rfunc1 = getattr(ps.rfunc, rfunc_name)(cutoff=0.5)

    # Create the exact same instance using to_dict
    data = rfunc1.to_dict()
    rfunc_class = data.pop("class")  # Determine response class
    rfunc_up = data.pop("up", None)
    rfunc_gsf = data.pop("gain_scale_factor", None)
    rfunc2 = getattr(ps.rfunc, rfunc_class)(**data)
    rfunc2.update_rfunc_settings(up=rfunc_up, gain_scale_factor=rfunc_gsf)

    if rfunc_name == "HantushWellModel":
        rfunc1.set_distances(100.0)
        rfunc2.set_distances(100.0)

    p1 = rfunc1.get_init_parameters("test").initial.to_numpy()
    p2 = rfunc2.get_init_parameters("test").initial.to_numpy()

    assert (rfunc1.step(p1) - rfunc2.step(p2)).sum() == 0.0



"""This module contains methods to compute the Standardized Groundwater Index."""

from numpy import linspace
from pandas import DataFrame, Series
from scipy.stats import norm


def sgi(series: Series) -> Series:
    """Method to compute the Standardized Groundwater Index
    :cite:t:`bloomfield_analysis_2013`.

    Parameters
    ----------
    series: pandas.Series or Pandas.DataFrame

    Returns
    -------
    sgi_series: pandas.Series or Pandas.DataFrame
        Pandas time series of the groundwater levels. Time series index should be a
        pandas DatetimeIndex.

    Notes
    -----
    The Standardized Groundwater Index (SGI) is a non-parametric method to
    standardize groundwater levels. The SGI is calculated for each month
    separately. The SGI is a dimensionless index and is used to compare
    groundwater levels across different wells. It is generally recommended to resample
    the data to a monthly time series before computing the SGI.

    """
    if isinstance(series, DataFrame):
        series = series.apply(sgi)
    elif isinstance(series, Series):
        series = series.dropna().copy()  # Create a copy to ensure series is untouched.

        # Loop over the months
        for month in range(1, 13):
            data = series[series.index.month == month]
            n = data.size  # Number of observations
            pmin = 1 / (2 * n)
            pmax = 1 - pmin
            sgi_values = norm.ppf(linspace(pmin, pmax, n))
            series.loc[data.sort_values().index] = sgi_values

    return series



import numpy as np
import pandas as pd

import pastas as ps


class TestGXG(object):
    def test_ghg(self) -> None:
        idx = pd.to_datetime(["20160114", "20160115", "20160128", "20160214"])
        s = pd.Series([10.0, 3.0, 30.0, 20.0], index=idx)
        v = ps.stats.ghg(s, min_n_meas=1, min_n_years=1)
        assert v == 30.0

    def test_ghg_ffill(self) -> None:
        idx = pd.to_datetime(["20160101", "20160115", "20160130"])
        s = pd.Series([0.0, 0.0, 10.0], index=idx)
        v = ps.stats.ghg(s, fill_method="ffill", limit=15, min_n_meas=1, min_n_years=1)
        assert v == 0.0

    def test_ghg_bfill(self) -> None:
        idx = pd.to_datetime(["20160101", "20160115", "20160130"])
        s = pd.Series([0.0, 0.0, 10.0], index=idx)
        v = ps.stats.ghg(s, fill_method="bfill", limit=15, min_n_meas=1, min_n_years=1)
        # TODO is this correct?
        assert v == 10.0

    def test_ghg_linear(self) -> None:
        idx = pd.to_datetime(["20160101", "20160110", "20160120", "20160130"])
        s = pd.Series([0.0, 0.0, 10.0, 10.0], index=idx)
        v = ps.stats.ghg(s, fill_method="linear", min_n_meas=1, min_n_years=1, limit=8)
        # TODO is this correct?
        assert v == 10.0

    def test_ghg_len_yearly(self) -> None:
        idx = pd.date_range("20000101", "20550101", freq="d")
        s = pd.Series(np.ones(len(idx)), index=idx)
        v = ps.stats.ghg(s, output="yearly")
        assert v.notna().sum() == 55

    def test_glg(self) -> None:
        idx = pd.date_range("20000101", "20550101", freq="d")
        s = pd.Series(
            [x.month + x.day for x in idx],
            index=idx,
        )
        v = ps.stats.glg(s, year_offset="YE")
        assert v == 16.0

    def test_glg_fill_limit(self) -> None:
        idx = pd.to_datetime(["20170115", "20170130", "20200101"])
        s = pd.Series(np.ones(len(idx)), index=idx)
        v = ps.stats.glg(
            s,
            fill_method="linear",
            limit=15,
            output="yearly",
            year_offset="YE",
            min_n_meas=1,
        )
        assert v.notna().sum() == 2

    def test_glg_fill_limit_null(self) -> None:
        idx = pd.to_datetime(["20170101", "20170131", "20200101"])
        s = pd.Series(np.ones(len(idx)), index=idx)
        v = ps.stats.glg(
            s,
            fill_method="linear",
            limit=None,
            output="yearly",
            year_offset="YE",
            min_n_meas=1,
        )
        assert v.notna().sum() == 3

    def test_gvg(self) -> None:
        idx = pd.to_datetime(["20170314", "20170328", "20170414", "20170428"])
        s = pd.Series([1.0, 2.0, 3.0, 4], index=idx)
        v = ps.stats.gvg(
            s, fill_method="linear", output="mean", min_n_meas=1, min_n_years=1
        )
        assert v == 2.0

    def test_gvg_nan(self) -> None:
        idx = pd.to_datetime(["20170228", "20170428", "20170429"])
        s = pd.Series([1.0, 2.0, 3.0], index=idx)
        v = ps.stats.gvg(
            s, fill_method=None, output="mean", min_n_meas=1, min_n_years=1
        )
        assert np.isnan(v)

        # def test_gxg_series(self) -> None:
        #     s = pd.read_csv('data\\hseries_gxg.csv', index_col=0, header=0,
        #                     parse_dates=True, dayfirst=True).squeeze()
        #     ps = Model(s)
        #     ps.freq = 'D'
        #     ghg = ps.stats.ghg(s)
        #     glg = ps.stats.glg(s)
        #     gvg = ps.stats.gvg(s)
        #     print('\n')
        #     print('calculated GXG\'s classic method: \n')
        #     print(('GHG: {ghg:.2f} m+NAP\n'
        #            'GLG: {glg:.2f} m+NAP\n'
        #            'GVG: {gvg:.2f} m+NAP\n').format(
        #         ghg=ghg, glg=glg, gvg=gvg))
        #     print('Menyanthes GXG\'s: \n')
        #     print(('GHG: {ghg:.2f} m+NAP\n'
        #            'GLG: {glg:.2f} m+NAP\n'
        #            'GVG: {gvg:.2f} m+NAP\n').format(
        #         ghg=-3.23, glg=-3.82, gvg=-3.43))

        # def test_gxg_series(self, capsys) -> None:
        #     s = pd.read_csv(r'data/hseries_gxg.csv', index_col=0, header=0,
        #         parse_dates=True, dayfirst=True).squeeze()
        #     ps = Model(s)
        #     ghg = ps.stats.ghg(s)
        #     glg = ps.stats.glg(s)
        #     gvg = ps.stats.gvg(s)
        #     with capsys.disabled():
        #         print('\n')
        #         print('calculated GXG\'s: \n')
        #         print(('GHG: {ghg:.2f} m+NAP\n'
        #                'GLG: {glg:.2f} m+NAP\n'
        #                'GVG: {gvg:.2f} m+NAP\n').format(
        #                ghg=ghg, glg=glg, gvg=gvg))
        #         print('Menyanthes GXG\'s: \n')
        #         print(('GHG: {ghg:.2f} m+NAP\n'
        #                'GLG: {glg:.2f} m+NAP\n'
        #                'GVG: {gvg:.2f} m+NAP\n').format(
        #                ghg=-3.23, glg=-3.82, gvg=-3.43))



"""This module contains the objective functions that can be used with the pastas
`EmceeSolve` solver.

"""

from numpy import log, pi
from pandas import DataFrame


class GaussianLikelihood:
    """Gaussian likelihood function.

    Notes
    -----
    The Gaussian log-likelihood function is defined as:

    .. math::
        \\log(L) = -\\frac{N}{2}\\log(2\\pi\\sigma^2) +
        \\frac{\\sum_{i=1}^N - \\epsilon_i^2}{2\\sigma^2}

    where :math:`N` is the number of observations, :math:`\\sigma^2` is the variance of
    the residuals, and :math:`\\epsilon_i` is the residual at time :math:`i`. The
    parameter :math:`\\sigma^2` need to be estimated.

    """

    _name = "GaussianLikelihood"

    def __init__(self):
        self.nparam = 1

    def get_init_parameters(self, name: str) -> DataFrame:
        """Get the initial parameters for the log-likelihood function.

        Parameters
        ----------
        name: str
            Name of the log-likelihood function.

        Returns
        -------
        parameters: DataFrame
            Initial parameters for the log-likelihood function.

        """
        parameters = DataFrame(
            columns=["initial", "pmin", "pmax", "vary", "stderr", "name", "dist"]
        )
        parameters.loc[name + "_sigma"] = (0.05, 1e-10, 1, True, 0.01, name, "uniform")
        return parameters

    def compute(self, rv, p):
        """Compute the log-likelihood.

        Parameters
        ----------
        rv: array
            Residuals of the model.
        p: array or list
            Parameters of the log-likelihood function.

        Returns
        -------
        ln: float
            Log-likelihood

        """
        sigma = p[-1]
        N = rv.size
        ln = -0.5 * N * log(2 * pi * sigma) + sum(-(rv**2) / (2 * sigma))
        return ln


class GaussianLikelihoodAr1:
    """Gaussian likelihood function with AR1 autocorrelated residuals.

    Notes
    -----
    The Gaussian log-likelihood function with AR1 autocorrelated residual is defined as:

    .. math::
        \\log(L) = -\\frac{N-1}{2}\\log(2\\pi\\sigma^2) +
        \\frac{\\sum_{i=1}^N - (\\epsilon_i - \\phi \\epsilon_{i-\\Delta t})^2}
        {2\\sigma^2}

    where :math:`N` is the number of observations, :math:`\\sigma^2` is the
    variance of the residuals, :math:`\\epsilon_i` is the residual at time
    :math:`i` and :math:`\\mu` is the mean of the residuals. :math:`\\Delta t` is
    the time step between the observations. :math:`\\phi` is the autoregressive
    parameter. The parameters :math:`\\phi` and :math:`\\sigma^2` need to be estimated.

    """

    _name = "GaussianLikelihoodAr1"

    def __init__(self):
        self.nparam = 2

    def get_init_parameters(self, name: str) -> DataFrame:
        """Get the initial parameters for the log-likelihood function.

        Parameters
        ----------
        name: str
            Name of the log-likelihood function.

        Returns
        -------
        parameters: DataFrame
            Initial parameters for the log-likelihood function.

        """
        parameters = DataFrame(
            columns=["initial", "pmin", "pmax", "vary", "stderr", "name", "dist"]
        )
        parameters.loc[name + "_sigma"] = (0.05, 1e-10, 1, True, 0.01, name, "uniform")
        parameters.loc[name + "_theta"] = (
            0.5,
            1e-10,
            0.99999,
            True,
            0.2,
            name,
            "uniform",
        )
        return parameters

    def compute(self, rv, p):
        """Compute the log-likelihood.

        Parameters
        ----------
        rv: array
            Residuals of the model.
        p: array or list
            Parameters of the log-likelihood function.

        Returns
        -------
        ln: float
            Log-likelihood.

        """
        sigma = p[-2]
        theta = p[-1]
        N = rv.size
        ln = -(N - 1) / 2 * log(2 * pi * sigma) + sum(
            -((rv[1:] - theta * rv[0:-1]) ** 2) / (2 * sigma)
        )
        return ln
